@@ -36,7 +36,7 @@
 #include <libgda/gda-util.h>
 #ifdef HAVE_FAM
 #include <fam.h>
-#include <glib/giochannel.h>
+#include <glib.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <unistd.h>
@@ -140,7 +140,7 @@ gda_config_class_init (GdaConfigClass *klass)
                               G_SIGNAL_RUN_FIRST,
                               G_STRUCT_OFFSET (GdaConfigClass, dsn_added),
                               NULL, NULL,
-                              gda_marshal_VOID__POINTER,
+                              _gda_marshal_VOID__POINTER,
                               G_TYPE_NONE, 1, G_TYPE_POINTER);
 	gda_config_signals[DSN_TO_BE_REMOVED] =
                 g_signal_new ("dsn-to-be-removed",
@@ -148,7 +148,7 @@ gda_config_class_init (GdaConfigClass *klass)
                               G_SIGNAL_RUN_FIRST,
                               G_STRUCT_OFFSET (GdaConfigClass, dsn_to_be_removed),
                               NULL, NULL,
-                              gda_marshal_VOID__POINTER,
+                              _gda_marshal_VOID__POINTER,
                               G_TYPE_NONE, 1, G_TYPE_POINTER);
 	gda_config_signals[DSN_REMOVED] =
                 g_signal_new ("dsn-removed",
@@ -156,7 +156,7 @@ gda_config_class_init (GdaConfigClass *klass)
                               G_SIGNAL_RUN_FIRST,
                               G_STRUCT_OFFSET (GdaConfigClass, dsn_removed),
                               NULL, NULL,
-                              gda_marshal_VOID__POINTER,
+                              _gda_marshal_VOID__POINTER,
                               G_TYPE_NONE, 1, G_TYPE_POINTER);
 	gda_config_signals[DSN_CHANGED] =
                 g_signal_new ("dsn-changed",
@@ -164,7 +164,7 @@ gda_config_class_init (GdaConfigClass *klass)
                               G_SIGNAL_RUN_FIRST,
                               G_STRUCT_OFFSET (GdaConfigClass, dsn_changed),
                               NULL, NULL,
-                              gda_marshal_VOID__POINTER,
+                              _gda_marshal_VOID__POINTER,
                               G_TYPE_NONE, 1, G_TYPE_POINTER);
 
 	/* Properties */
@@ -765,7 +765,7 @@ gda_config_define_dsn (const GdaDsnInfo *info, GError **error)
 
 	if (info->is_system && !unique_instance->priv->system_config_allowed) {
 		g_set_error (error, GDA_CONFIG_ERROR, GDA_CONFIG_PERMISSION_ERROR,
-			     _("Can't manage system-wide configuration"));
+			      "%s", _("Can't manage system-wide configuration"));
 		GDA_CONFIG_UNLOCK ();
 		return FALSE;
 	}
@@ -854,7 +854,7 @@ gda_config_remove_dsn (const gchar *dsn_name, GError **error)
 	}
 	if (info->is_system && !unique_instance->priv->system_config_allowed) {
 		g_set_error (error, GDA_CONFIG_ERROR, GDA_CONFIG_PERMISSION_ERROR,
-			     _("Can't manage system-wide configuration"));
+			      "%s", _("Can't manage system-wide configuration"));
 		GDA_CONFIG_UNLOCK ();
 		return FALSE;
 	}
@@ -879,7 +879,7 @@ gda_config_remove_dsn (const gchar *dsn_name, GError **error)
 }
 
 /**
- * gda_config_dsn_needs_auth
+ * gda_config_dsn_needs_authentication
  * @dsn_name: the name of a DSN, in the "[&lt;username&gt;[:&lt;password&gt;]@]&lt;DSN&gt;" format
  * 
  * Tells if the data source identified as @dsn_name needs any authentication. If a &lt;username&gt;
@@ -942,7 +942,7 @@ gda_config_list_dsn (void)
  *
  * Get the number of defined DSN
  *
- * Return: the number of defined DSN
+ * Returns: the number of defined DSN
  */
 gint
 gda_config_get_nb_dsn (void)
@@ -959,7 +959,7 @@ gda_config_get_nb_dsn (void)
 
 /**
  * gda_config_get_dsn_info_index
- * @dsn_name:
+ * @dsn_name: a DSN
  * 
  * Get the index (starting at 0) of the DSN named @dsn_name
  * 
@@ -986,7 +986,7 @@ gda_config_get_dsn_info_index (const gchar *dsn_name)
 
 /**
  * gda_config_get_dsn_info_at_index
- * @index:
+ * @index: an index
  *
  * Get a pointer to a read-only #GdaDsnInfo at the @index position
  *
@@ -1026,10 +1026,9 @@ gda_config_can_modify_system_config (void)
 
 /**
  * gda_config_get_provider_info
- * @provider_name:
+ * @provider_name: a database provider
  *
  * Get some information about the a database provider (adaptator) named 
- * @provider_name
  *
  * Returns: a pointer to read-only #GdaProviderInfo structure, or %NULL if not found
  */
@@ -1061,7 +1060,7 @@ gda_config_get_provider_info (const gchar *provider_name)
 
 /**
  * gda_config_get_provider
- * @provider_name:
+ * @provider_name: a database provider
  * @error: a place to store errors, or %NULL
  *
  * Get a pointer to the session-wide #GdaServerProvider for the
@@ -1075,6 +1074,7 @@ gda_config_get_provider (const gchar *provider_name, GError **error)
 {
 	InternalProvider *ip;
 	GdaServerProvider  *(*plugin_create_provider) (void);
+	GdaServerProvider  *(*plugin_create_sub_provider) (const gchar *provider_name);
 
 	g_return_val_if_fail (provider_name, NULL);
 	GDA_CONFIG_LOCK ();
@@ -1091,15 +1091,32 @@ gda_config_get_provider (const gchar *provider_name, GError **error)
 	}
 	
 	/* need to actually create the provider object */
-	g_assert (ip->handle);
-	g_module_symbol (ip->handle, "plugin_create_provider", (gpointer) &plugin_create_provider);
-	if (!plugin_create_provider) {
-		g_set_error (error, GDA_CONFIG_ERROR, GDA_CONFIG_PROVIDER_CREATION_ERROR,
-			     _("Can't instantiate provider '%s'"), provider_name);
-		GDA_CONFIG_UNLOCK ();
-		return NULL;
+	if (!ip->handle) {
+		GdaProviderInfo *info = (GdaProviderInfo*) ip;
+		ip->handle = g_module_open (info->location, G_MODULE_BIND_LAZY);
+		if (!ip->handle) {
+			g_set_error (error, GDA_CONFIG_ERROR, GDA_CONFIG_PROVIDER_CREATION_ERROR,
+				     _("Can't load provider: %s"), g_module_error ());
+			return NULL;
+		}
+
+		void (*plugin_init) (const gchar *);
+		if (g_module_symbol (ip->handle, "plugin_init", (gpointer *) &plugin_init)) {
+			gchar *dirname = g_path_get_dirname (info->location);
+			plugin_init (dirname);
+			g_free (dirname);
+		}
 	}
-	ip->instance = plugin_create_provider ();
+
+	g_module_symbol (ip->handle, "plugin_create_provider", (gpointer) &plugin_create_provider);
+	if (plugin_create_provider)
+		ip->instance = plugin_create_provider ();
+	else {
+		g_module_symbol (ip->handle, "plugin_create_sub_provider", (gpointer) &plugin_create_sub_provider);
+		if (plugin_create_sub_provider)
+			ip->instance = plugin_create_sub_provider (provider_name);
+	}
+	
 	if (!ip->instance) {
 		g_set_error (error, GDA_CONFIG_ERROR, GDA_CONFIG_PROVIDER_CREATION_ERROR,
 			     _("Can't instantiate provider '%s'"), provider_name);
@@ -1164,7 +1181,10 @@ gda_config_list_providers (void)
 		gda_data_model_set_value_at (model, 0, row, value, NULL);
 		gda_value_free (value);
 
-		value = gda_value_new_from_string (info->description, G_TYPE_STRING);
+		if (info->description) 
+			value = gda_value_new_from_string (info->description, G_TYPE_STRING);
+		else
+			value = gda_value_new_null ();
 		gda_data_model_set_value_at (model, 1, row, value, NULL);
 		gda_value_free (value);
 
@@ -1235,6 +1255,82 @@ load_all_providers (void)
 	}
 }
 
+static InternalProvider *
+create_internal_provider (const gchar *path,
+			  const gchar *prov_name, const gchar *prov_descr,
+			  gchar *dsn_spec, gchar *auth_spec)
+{
+	InternalProvider *ip;
+	GdaProviderInfo *info;
+
+	ip = g_new0 (InternalProvider, 1);
+	ip->handle = NULL;
+	info = (GdaProviderInfo*) ip;
+	info->location = g_strdup (path);
+
+	info->id = g_strdup (prov_name);
+	info->description = g_strdup (prov_descr);
+
+	/* DSN parameters */
+	info->dsn_params = NULL;
+	if (dsn_spec) {
+		GError *error = NULL;
+		info->dsn_params = gda_set_new_from_spec_string (dsn_spec, &error);
+		if (!info->dsn_params) {
+			g_warning ("Invalid format for provider '%s' DSN spec : %s",
+				   info->id,
+				   error ? error->message : "Unknown error");
+			if (error)
+				g_error_free (error);
+		}
+		if (!info->dsn_params) {
+			/* there may be traces of the provider installed but some parts are missing,
+			   forget about that provider... */
+			internal_provider_free (ip);
+			g_free (dsn_spec);
+			return NULL;
+		}
+		g_free (dsn_spec);
+	}
+	else
+		g_warning ("Provider '%s' does not provide a DSN spec", info->id);
+
+	/* Authentication parameters */
+	info->auth_params = NULL;
+	if (auth_spec) {
+		GError *error = NULL;
+
+		info->auth_params = gda_set_new_from_spec_string (auth_spec, &error);
+		if (!info->auth_params) {
+			g_warning ("Invalid format for provider '%s' AUTH spec : %s",
+				   info->id,
+				   error ? error->message : "Unknown error");
+			if (error)
+				g_error_free (error);
+		}
+
+		if (!info->auth_params) {
+			/* there may be traces of the provider installed but some parts are missing,
+			   forget about that provider... */
+			internal_provider_free (ip);
+			g_free (auth_spec);
+			return NULL;
+		}
+		g_free (auth_spec);
+	}
+	else {
+		/* default to username/password */
+		GdaHolder *h;
+		info->auth_params = gda_set_new_inline (2, "USERNAME", G_TYPE_STRING, NULL,
+							"PASSWORD", G_TYPE_STRING, NULL);
+		h = gda_set_get_holder (info->auth_params, "USERNAME");
+		g_object_set (G_OBJECT (h), "name", _("Username"), NULL);
+		h = gda_set_get_holder (info->auth_params, "PASSWORD");
+		g_object_set (G_OBJECT (h), "name", _("Password"), NULL);
+	}
+	return ip;
+}
+
 static void 
 load_providers_from_dir (const gchar *dirname, gboolean recurs)
 {
@@ -1252,17 +1348,25 @@ load_providers_from_dir (const gchar *dirname, gboolean recurs)
 		g_error_free (err);
 		return;
 	}
-
+	
 	while ((name = g_dir_read_name (dir))) {
-		InternalProvider *ip;
-		GdaProviderInfo *info;
 		GModule *handle;
 		gchar *path;
+
+		/* initialization method */
 		void (*plugin_init) (const gchar *);
+
+		/* methods for shared libraries which provide only one type of provider */
 		const gchar * (* plugin_get_name) (void);
 		const gchar * (* plugin_get_description) (void);
 		gchar * (* plugin_get_dsn_spec) (void);
 		gchar * (* plugin_get_auth_spec) (void);
+		
+		/* methods for shared libraries which provide several types of providers (ODBC, JDBC, ...) */
+		const gchar ** (* plugin_get_sub_names) (void);
+		const gchar * (* plugin_get_sub_description) (const gchar *name);
+		gchar * (* plugin_get_sub_dsn_spec) (const gchar *name);
+		gchar * (* plugin_get_sub_auth_spec) (const gchar *name);
 
 		if (recurs) {
 			gchar *cname;
@@ -1290,106 +1394,76 @@ load_providers_from_dir (const gchar *dirname, gboolean recurs)
 			continue;
 		}
 
-		if (g_module_symbol (handle, "plugin_init", (gpointer *) &plugin_init))
+		if (g_module_symbol (handle, "plugin_init", (gpointer *) &plugin_init)) {
 			plugin_init (dirname);
-		g_module_symbol (handle, "plugin_get_name", (gpointer *) &plugin_get_name);
-		g_module_symbol (handle, "plugin_get_description", (gpointer *) &plugin_get_description);
-		g_module_symbol (handle, "plugin_get_dsn_spec", (gpointer *) &plugin_get_dsn_spec);
-		g_module_symbol (handle, "plugin_get_auth_spec", (gpointer *) &plugin_get_auth_spec);
-
-		ip = g_new0 (InternalProvider, 1);
-		ip->handle = handle;
-		info = (GdaProviderInfo*) ip;
-		info->location = path;
-
-		if (plugin_get_name != NULL)
-			info->id = g_strdup (plugin_get_name ());
-		else
-			info->id = g_strdup (name);
-
-		if (plugin_get_description != NULL)
-			info->description = g_strdup (plugin_get_description ());
-		else
-			info->description = NULL;
-
-		/* DSN parameters */
-		info->dsn_params = NULL;
-		if (plugin_get_dsn_spec) {
-			GError *error = NULL;
-			gchar *dsn_spec;
-
-			dsn_spec = plugin_get_dsn_spec ();
-			if (dsn_spec) {
-				info->dsn_params = gda_set_new_from_spec_string (dsn_spec, &error);
-				if (!info->dsn_params) {
-					g_warning ("Invalid format for provider '%s' DSN spec : %s",
-						   info->id,
-						   error ? error->message : "Unknown error");
-					if (error)
-						g_error_free (error);
-				}
-				g_free (dsn_spec);
-			}
-			if (!info->dsn_params) {
-				/* there may be traces of the provider installed but some parts are missing,
-				   forget about that provider... */
-				internal_provider_free (ip);
-				ip = NULL;
-				continue;
-			}
 		}
-		else
-			g_warning ("Provider '%s' does not provide a DSN spec", info->id);
+		else {
+			g_module_close (handle);
+			continue;
+		}
+		g_module_symbol (handle, "plugin_get_name",
+				 (gpointer *) &plugin_get_name);
+		g_module_symbol (handle, "plugin_get_description",
+				 (gpointer *) &plugin_get_description);
+		g_module_symbol (handle, "plugin_get_dsn_spec",
+				 (gpointer *) &plugin_get_dsn_spec);
+		g_module_symbol (handle, "plugin_get_auth_spec",
+				 (gpointer *) &plugin_get_auth_spec);
+		g_module_symbol (handle, "plugin_get_sub_names",
+				 (gpointer *) &plugin_get_sub_names);
+		g_module_symbol (handle, "plugin_get_sub_description",
+				 (gpointer *) &plugin_get_sub_description);
+		g_module_symbol (handle, "plugin_get_sub_dsn_spec",
+				 (gpointer *) &plugin_get_sub_dsn_spec);
+		g_module_symbol (handle, "plugin_get_sub_auth_spec",
+				 (gpointer *) &plugin_get_sub_auth_spec);
 
-		/* Authentication parameters */
-		info->auth_params = NULL;
-		if (plugin_get_auth_spec) {
-			GError *error = NULL;
-			gchar *auth_spec;
-
-			auth_spec = plugin_get_auth_spec ();
-			if (auth_spec) {
-				info->auth_params = gda_set_new_from_spec_string (auth_spec, &error);
-				if (!info->auth_params) {
-					g_warning ("Invalid format for provider '%s' AUTH spec : %s",
-						   info->id,
-						   error ? error->message : "Unknown error");
-					if (error)
-						g_error_free (error);
+		if (plugin_get_sub_names) {
+			const gchar **subnames = plugin_get_sub_names ();
+			const gchar **ptr;
+			for (ptr = subnames; ptr && *ptr; ptr++) {
+				InternalProvider *ip;
+								
+				ip = create_internal_provider (path, *ptr,
+							       plugin_get_sub_description ? 
+							       plugin_get_sub_description (*ptr) : NULL,
+							       plugin_get_sub_dsn_spec ? 
+							       plugin_get_sub_dsn_spec (*ptr) : NULL,
+							       plugin_get_sub_auth_spec ?
+							       plugin_get_sub_auth_spec (*ptr) : NULL);
+				if (ip) {
+					unique_instance->priv->prov_list =
+						g_slist_prepend (unique_instance->priv->prov_list, ip);
+#ifdef GDA_DEBUG_NO
+					g_print ("Loaded '%s' sub-provider\n", ((GdaProviderInfo*) ip)->id);
+#endif
 				}
-				g_free (auth_spec);
-			}
-			if (!info->auth_params) {
-				/* there may be traces of the provider installed but some parts are missing,
-				   forget about that provider... */
-				internal_provider_free (ip);
-				ip = NULL;
-				continue;
 			}
 		}
 		else {
-			/* default to username/password */
-			GdaHolder *h;
-			info->auth_params = gda_set_new_inline (2, "USERNAME", G_TYPE_STRING, NULL,
-								"PASSWORD", G_TYPE_STRING, NULL);
-			h = gda_set_get_holder (info->auth_params, "USERNAME");
-			g_object_set (G_OBJECT (h), "name", _("Username"), NULL);
-			h = gda_set_get_holder (info->auth_params, "PASSWORD");
-			g_object_set (G_OBJECT (h), "name", _("Password"), NULL);
-		}
-
-		if (ip) {
-			unique_instance->priv->prov_list = g_slist_prepend (unique_instance->priv->prov_list, ip);
+			InternalProvider *ip;
+			ip = create_internal_provider (path,
+						       plugin_get_name ? plugin_get_name () : name,
+						       plugin_get_description ? plugin_get_description () : NULL,
+						       plugin_get_dsn_spec ? plugin_get_dsn_spec () : NULL,
+						       plugin_get_auth_spec ? plugin_get_auth_spec () : NULL);
+			if (ip) {
+				unique_instance->priv->prov_list =
+					g_slist_prepend (unique_instance->priv->prov_list, ip);
 #ifdef GDA_DEBUG_NO
-			g_print ("Loaded '%s' provider\n", info->id);
+				g_print ("Loaded '%s' provider\n", ((GdaProviderInfo*) ip)->id);
 #endif
+			}
 		}
+		g_module_close (handle);
 	}
 
 	/* free memory */
 	g_dir_close (dir);
 }
 
+
+
 /* sorting function */
 static gint
 data_source_info_compare (GdaDsnInfo *infoa, GdaDsnInfo *infob)
