@@ -25,6 +25,7 @@
 #include <libgda/gda-data-model.h>
 #include <libgda/gda-meta-store.h>
 #include <libgda/gda-decl.h>
+#include <libgda/gda-attributes-manager.h>
 
 G_BEGIN_DECLS
 
@@ -113,12 +114,10 @@ typedef struct {
 } GdaMetaView;
 
 /*
- * Struture to hold information about each object
- * which can be created in the internal GdaMetaStore's connection.
- * It is available for tables, views, triggers, ...
+ * Struture to hold information about each database object (tables, views, triggers, ...)
  *
  * Note: @obj_catalog, @obj_schema, @obj_name, @obj_short_name and @obj_full_name are case sensitive:
- *       one must call _identifier_needs_quotes() to know is it is necessary to surround by double quotes
+ *       one must call gda_sql_identifier_needs_quotes() to know if is it is necessary to surround by double quotes
  *       before using in an SQL statement
  */
 typedef struct {
@@ -146,10 +145,21 @@ typedef struct {
 	gchar        *column_type;
 	GType         gtype;
 	gboolean      pkey;
-        gboolean      nullok;
+	gboolean      nullok;
 	gchar        *default_value;
 } GdaMetaTableColumn;
 #define GDA_META_TABLE_COLUMN(x) ((GdaMetaTableColumn*)(x))
+const GValue *gda_meta_table_column_get_attribute (GdaMetaTableColumn *tcol, const gchar *att_name);
+void          gda_meta_table_column_set_attribute (GdaMetaTableColumn *tcol, const gchar *att_name, const GValue *value,
+						   GDestroyNotify destroy);
+/**
+ * gda_meta_table_column_set_attribute_static
+ *
+ * This function is similar to gda_meta_table_column_set_attribute() but for static strings
+ */
+#define gda_meta_table_column_set_attribute_static(holder,attribute,value) gda_meta_table_column_set_attribute((holder),(attribute),(value),NULL)
+
+void          gda_meta_table_column_foreach_attribute (GdaMetaTableColumn *tcol, GdaAttributesManagerFunc func, gpointer data);
 
 typedef struct {
 	GdaMetaDbObject  *meta_table;
@@ -172,6 +182,7 @@ GdaMetaDbObject    *gda_meta_struct_complement         (GdaMetaStruct *mstruct,
 gboolean            gda_meta_struct_complement_schema  (GdaMetaStruct *mstruct,
 							const GValue *catalog, const GValue *schema, GError **error);
 gboolean            gda_meta_struct_complement_default (GdaMetaStruct *mstruct, GError **error);
+gboolean            gda_meta_struct_complement_all     (GdaMetaStruct *mstruct, GError **error);
 gboolean            gda_meta_struct_complement_depend  (GdaMetaStruct *mstruct, GdaMetaDbObject *dbo,
 							GError **error);
 
